import 'dart:async';

import 'package:flutter/services.dart';

class XinPlatform {
  static const MethodChannel _channel = MethodChannel('xinlake_platform');

  /// Returns a list contains files picked, an empty list if the action was canceled,
  /// null if error occurs
  /// * [multiSelection]: Multiple selection
  /// ***
  /// Android. READ_EXTERNAL_STORAGE permission is required.
  /// * [mimeType]: Acceptable MIME types such as "image/jpeg", "audio/*", "application/xyz"
  /// * [cacheDir]:
  /// * [cacheOverwrite]: This parameter will be ignored if cacheDir is AndroidCacheDir.none
  /// ***
  /// Windows.
  /// * [openPath]: Path to open
  /// * [defaultPath]: Default open folder for the dialog, if openPath is set
  /// then defaultPath will be ignored.
  /// * [filterName]: Friendly name of the filter such as "JPEG Image"
  /// * [filterPattern]: Filter pattern such as "*.jpg; *.jpeg"
  static Future<List<String>?> pickFiles({
    bool multiSelection = false,
    // android
    String mimeType = "*/*",
    XinAndroidAppDir? cacheDir,
    bool cacheOverwrite = false,
    // windows
    String? openPath,
    String? defaultPath,
    required String filterName,
    required String filterPattern,
  }) async {
    try {
      return await _channel.invokeListMethod('pickFiles', {
        "multiSelection": multiSelection,
        "mimeType": mimeType,
        "cacheDirIndex": cacheDir?.index,
        "cacheOverwrite": cacheOverwrite,
        "openPath": openPath ?? "",
        "defaultPath": defaultPath ?? "",
        "filterName": filterName,
        "filterPattern": filterPattern,
      });
    } catch (error) {
      return null;
    }
  }

  static Future<void> setUiMode(
    XinUiMode mode, {
    int darkColor = 0xff101010,
    int lightColor = 0xffffffff,
    int animateMs = 100,
  }) async {
    try {
      return await _channel.invokeMethod('setUiMode', {
        "modeIndex": mode.index,
        "darkColor": darkColor,
        "lightColor": lightColor,
        "animateMs": animateMs,
      });
    } catch (error) {
      // ignored
    }
  }

  /// Get the absolute path of the app directory
  static Future<String?> getAppDir({
    XinAndroidAppDir? androidAppDir,
  }) async {
    try {
      return await _channel.invokeMethod("getAppDir", {
        "appDirIndex": androidAppDir?.index,
      });
    } catch (error) {
      return null;
    }
  }

  /// Get the app version info
  static Future<XinVersionInfo?> getAppVersion() async {
    try {
      final map = await _channel.invokeMapMethod<String, dynamic>("getAppVersion");
      if (map != null) {
        return XinVersionInfo(
          map["version"],
          map["updated-utc"],
        );
      }
    } catch (error) {
      // ignored
    }

    return null;
  }
}

enum XinUiMode { light, dark, system, custom }
enum XinAndroidAppDir { internalCache, internalFiles, externalCache, externalFiles }

class XinVersionInfo {
  String version;
  int updatedMsUtc;
  DateTime updatedTime;

  XinVersionInfo(this.version, this.updatedMsUtc)
      : updatedTime = DateTime.utc(1970, 1, 1, 0, 0, 0, updatedMsUtc, 0).toLocal();
}
